/*****************************************************************************
Temperatur-Messung mit LM75 und senden ueber RFM12-Funkmodul
Version:   1.0  14.10.2011

Prozessor: Atmega8 mit internem Takt ca. 1MHz
Quarz:     32768kHz Uhrenquarz fuer Timer2
Schaltung: funkthermometer.png

Funkmodul: RFM12B
Frequenz:  434MHz
Datenrate: 4.8kbps

Schaltung: tempsensor.png
------------------------
ATMEGA8           RFM12B (Funkmodul zum Senden und Empfangen)
------------------------
PB5               SCK
PB4               SDO
PB3               SDI
PB2               nSEL
PC2               nIRQ
PC0               FSK/DATA/NFFS
PC3               DCLK/CFIL/FFIT
PB0               nINT, VDI

------------------------
ATMEGA8           LM75A (es sind 2 angeschlossen, koennten aber bis 8 sein)
------------------------
PC4 (SDA)         1  (mit 4.7k Pullup)
PC5 (SCL)         2  (mit 4.7k Pullup)

PD1               LED1
PD0               LED2

PD2-PD7           frei fuer Erweiterung mit LCD-Modul


History:
1.10.2011     0.02 Erste Stromsparende Version
2.10.2011     0.03 Daten binaer senden, Waermemenge messen
6.10.2011     0.04 Uhrzeit auch mitsenden. Byteorder LSB first.
12.10.2011    0.05 Nach dem Senden auf Antwort warten und evtl Uhr neu setzen
14.10.2011    1.0  Erste funktionierende Version, aber ohne Antwort empfangen
              (alle Tests von 0.0-Versionen entfernt)

******************************************************************************/
#define XMINUTEN 1  //Es jede Minute gemessen aber nur alle X Minuten gesendet

//Quarz fuer Timer2 verwendet, CPU mit internem 1MHz:
#define F_CPU 1000000L

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>
#include "sensoren.h"

/********************************* RFM12 *********************************/
//Routinen zum Senden und Empfangen uebers RFM12B-Modul
#define PORT_SEL PORTB
#define PIN_SEL PINB
#define DDR_SEL DDRB
#define PORT_SDI PORTB
#define PIN_SDI PINB
#define DDR_SDI DDRB
#define PORT_SCK PORTB
#define PIN_SCK PINB
#define DDR_SCK DDRB
#define PORT_SDO PORTB
#define PIN_SDO PINB
#define DDR_SDO DDRB
#define PORT_IRQ PORTC
#define PIN_IRQ PINC
#define DDR_IRQ DDRC
#define PORT_DATA PORTC
#define PIN_DATA PINC
#define DDR_DATA DDRC
#define RFXX_SCK 5// |
#define RFXX_SDO 4// |RF_PORT
#define RFXX_SDI 3// |
#define RFXX_SEL 2// |
#define RF12_IRQ 2
#define RF12_DATA 0

#include "rfm12_modul.c"

/****************************** Zeitbehandlung *******************************/
static volatile uchar sec=0,min=0,stunden=0;
static volatile uint tage=0;

//Uhr nur mit Sekunden-Aufloesung, ohne Korrektur von Quarzungenauigkeiten:
#define F_QUARZ 32768
#define VORTEILER 128
const uchar NormalerCompwert=F_QUARZ/VORTEILER-1;

ISR(SIG_OUTPUT_COMPARE2)
{
 if(++sec==60)
   {sec=0;
    if(++min==60)
      {min=0;
       if(++stunden==24)
	 {stunden=0; ++tage;}
      }
   }
}

void long_delay(uint16_t ms)
{
 for(; ms>0; ms--) _delay_ms(1);
}

void timer2_init()
{
 ASSR  = (1<< AS2);              // Timer2 asynchron takten
 long_delay(1000);               // Einschwingzeit des 32kHz Quarzes
 TCCR2  = (1<<WGM21)|(1<<CS22)|(1<<CS20); // CTC-Modus, Vorteiler=128
 while((ASSR & (1<< TCR2UB)));   // Warte auf das Ende des Zugriffs
 TIFR   = (1<<OCF2);             // Interrupt lschen (*)
 TIMSK |= (1<<OCIE2);            // Timer Compare Match Interrupt freischalten
// (*) "Alternatively, OCF2 is cleared by writing a logic one to the flag."
 OCR2=NormalerCompwert;
 while((ASSR & (1<< OCR2UB)));   // Warte auf das Ende des Zugriffs
}

//Bei einem unerwarteten Interrupt nichts machen:
void __attribute__ ((naked)) __vector_default(void) {__asm__ ("reti");}

/************************* Temperatursensor LM75A ****************************/
#include "i2c.c"
#include "lm75a.c"

/***************************** Leuchtdioden **********************************/
#define PORT_LED PORTD
#define DDR_LED DDRD
#define LED1 1
#define LED2 0
#define LEDS_OUTPUT() DDR_LED|=((1<<LED1)|(1<<LED2))
#define LED1_OFF() PORT_LED&=~(1<<LED1)
#define LED1_ON()  PORT_LED|=(1<<LED1)
#define LED2_OFF() PORT_LED&=~(1<<LED2)
#define LED2_ON()  PORT_LED|=(1<<LED2)

/***************************** Hauptprogramm *********************************/
int main(void)
{
 unsigned char i;
 uchar altmin;
 int temp1,temp2; //Temperaturen in 8tel Grad Celsius
 long waermemenge=0;
 uchar sendeminute=min;
 datenblock block;
 LEDS_OUTPUT();

 timer2_init();
 sei(); //Interrupt einschalten
 twi_init(); //I2C initialisieren
 //lm75a_init(); //nur zum Solltemperatur setzen benoetigt

 for(i=0;i<3;i++) //POWER ON indication: LED blink 3 times
   {
    long_delay(200); LED1_ON(); LED2_ON();
    long_delay(200); LED1_OFF(); LED2_OFF();
   }

 rfm12_init();

 while(1) //Hauptschlaufe
   {
    LED1_ON(); //mit gruener LED anzeigen dass Temp gemessen wird
    setfuehlernummer(1); //entspricht der Einstellung mit A2,A1,A0 am LM75
    lm75a_nosleep(); //Fuehler 1 vom Schlafmodus aufwecken
    long_delay(1);
    setfuehlernummer(2);
    lm75a_nosleep(); //Fuehler 2 vom Schlafmodus aufwecken
    long_delay(60); //Warten bis die Sensoren Messungen gemacht haben (mindestens 50 Millisekunden noetig?)
    setfuehlernummer(1);
    temp1=lm75a_readtemp();
    long_delay(1);
    lm75a_sleep(); //Fuehler 1 wieder in Schlafmodus legen
    long_delay(1);
    setfuehlernummer(2);
    temp2=lm75a_readtemp();
    long_delay(1);
    lm75a_sleep(); //Fuehler 2 wieder in Schlafmodus legen
    LED1_OFF();

    //Um abgegebene Waermemenge eines Heizkoerpers zu messen:
    if(temp1>=temp2+4*8 &&                       //nur wenn Temperaturdifferenz mindestens 4 Grad
       temp1>=23*8 &&                            //und Heizkoerper mindestens 23 Grad
       temp1!=LM75A_ERROR && temp2!=LM75A_ERROR) //und keine Fehler aufgetreten sind.
     {
      waermemenge += temp1-temp2; //Temperaturdifferenz in Achtelgrad aufsummieren. Das Auswerteprogramm
                                  //muss dann noch Flaeche und K-Wert des Heizkoerpers beruecksichtigen.
     }

    if(min==sendeminute)
     {
      LED2_ON(); //mit roter LED anzeigen dass gesendet wird
      block.n=16; //Blocklaenge
      block.id=ID1;
      block.temp1=temp1;
      block.temp2=temp2;
      block.zaehler=waermemenge;
      block.sec=sec;
      block.min=min;
      block.stu=stunden;
      block.batt=3000; //provi. (Batteriespannung in Millivolt)
      rfm12_senden(block.data);
      LED2_OFF();
      sendeminute += XMINUTEN;
      if(sendeminute>=60) sendeminute -= 60;
     }
    //moegliche Schlafmoden: SLEEP_MODE_IDLE, _PWR_DOWN, _PWR_SAVE, _ADC, _STANDBY, _EXT_STANDBY
    set_sleep_mode(SLEEP_MODE_PWR_SAVE);
        {
	 altmin=min;
	 while(altmin==min) sleep_mode(); //naechste Minute abwarten
	}
   };//Ende Hauptschlaufe

 return 0;//wird nie erreicht.
}
